#ifndef _powerEW_
#define _powerEW_

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/* -------------------------- Prototypes ----------------------------- */

int TFmdm_set_cosm(float, float, float, int, float, float, float);
float TFmdm_onek_mpc(float);
float TFmdm_onek_hmpc(float);

#endif
